/*
 *  This file is part of Siena, a wide-area event notification system.
 *  See http://www.inf.usi.ch/carzaniga/siena/
 *  
 *  Author: Antonio Carzaniga
 *  See the file AUTHORS for full details. 
 *  
 *  Copyright (C) 1998-2003 University of Colorado
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA, or send email to one of the authors.
 *  
 *  
 *  $Id: sff_parser.yy,v 1.45 2010-03-11 15:16:41 carzanig Exp $
 */
%{
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <set>
#include <map>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#else
# error you need unistd.h
#endif

#include <siena/allocator.h>
#include <siena/forwarding.h>
#include <siena/fwdtable.h>
#include <siena/btable.h>
#include <siena/bxtable.h>
#include <siena/bctable.h>
#include <siena/bvtable.h>
#ifdef HAVE_CUDD
#include <siena/bddbtable.h>
#endif

#include "yysff.h"
#include "yysfftypes.h"

#include "range_set.h"
#include "simple_fwd_types.h"
#include "timing.h"

#include "statistics.h"

#define YYERROR_VERBOSE
#ifdef YYDEBUG
#undef YYDEBUG
#endif

bool errors = false;
int yyerror(const char *);

static siena::Timer sff_timer;
static siena::Timer parser_timer;
#ifndef WITH_STATISTICS
static siena::Timer match_timer;
#endif

static std::ofstream * sff_foutput = 0;
static std::ostream * sff_output = & std::cout;

class ICounter : public siena::MatchHandler {
public:
    ICounter() : c(0) { }
    virtual bool output(siena::if_t) throw();

    unsigned long long get_count() { return c; }
private:
    unsigned long long c;
};

bool ICounter::output(siena::if_t x) throw() {
    ++c;
    return false;
}

class IPrinter : public ICounter {
public:
    IPrinter(): ICounter() {};
    virtual bool output(siena::if_t) throw();
    void flush(std::ostream & os);

 private:
    std::set<siena::if_t> iset;
};

bool IPrinter::output(siena::if_t x) throw() {
    ICounter::output(x);
    iset.insert(x);
    return false;
}

void IPrinter::flush(std::ostream & os) {
    os << "->";
    std::set<siena::if_t>::const_iterator i;
    for(i = iset.begin(); i != iset.end(); ++i)
	os << ' ' << *i;
    os << std::endl;
}

 bool stats_only = false; 
      
 static sff_output_level_t output_level = SFF_VERBOSE; 

 unsigned long i_counter;
 unsigned long f_counter;
 unsigned long c_counter;
 unsigned long m_counter;
 unsigned long a_counter;
 unsigned long x_counter;
 unsigned long xx_counter;

 static const char * StatsFormat = 0;

 static siena::range_set<int> *ifconfig_ranges = NULL;

 static siena::ForwardingTable * FT = 0;

 static bool consolidate_guard = false;
 static siena::FTAllocator Mem;

 static siena::string_t new_sx_string(const string * x) {
     char * s = new (Mem) char[x->length()];
     char * t = s;
     for(string::const_iterator c = x->begin(); c != x->end(); ++c)
	 *t++ = *c;
     return siena::string_t(s, t);
 }

 static siena::string_t new_sx_string() {
     return siena::string_t(0, 0);
 }

 std::ostream & operator << (std::ostream & os,  const siena::constraint & a)
     throw() {

     siena::string_t name = a.name();
     for(const char * c = name.begin; c != name.end; ++c)
	 os << *c;

     os << ' ';
     switch(a.op()) {
     case siena::eq_id: os << "="; break;
     case siena::lt_id: os << "<"; break;
     case siena::gt_id: os << ">"; break;
     case siena::sf_id: os << "*="; break;
     case siena::pf_id: os << "=*"; break;
     case siena::ss_id: os << "**"; break;
     case siena::any_id: os << "any"; break;
     case siena::ne_id: os << "!="; break;
     case siena::re_id: os << "~="; break;
     }

     os << ' ';
     switch(a.type()) {
     case siena::string_id: {
	 siena::string_t value = a.string_value();
	 os << '"';
	 for(const char * c = value.begin; c != value.end; ++c) {
	     switch(*c) {
	     case '"': os << "\\\""; break;
	     case '\\': os << "\\\\"; break;
	     default: os << *c; break;
	     }
	 }
	 os << '"';
	 break;
     }
     case siena::int_id: os << a.int_value(); break;
     case siena::double_id: os << a.int_value(); break;
     case siena::bool_id: os << (a.bool_value() ? "true" : "false"); break;
     case siena::anytype_id: os << "any"; break;
     }
     return os;
 }

 std::ostream & operator << (std::ostream & os,  const siena::predicate & p)
     throw() {
     siena::predicate::iterator * pi;
     if ((pi = p.first())) {
	 bool first_filter = true;
	 do {
	     siena::filter::iterator * fi;
	     if ((fi = pi->first())) {
		 if (first_filter) {
		     first_filter = false;
		 } else {
		     os << "| ";
		 }
		 bool first_constraint = true;
		 do {
		     if (first_constraint) {
			 first_constraint = false;
		     } else {
			 os << ", ";
		     }
		     os << *fi;
		 } while (fi->next());
		 delete(fi);
		 os << std::endl;
	     }
	 } while (pi->next());
	 delete (pi);
     }
     return os;
 }

%}

%pure_parser

/*
 * Keywords & other lex tokens
 */
%token <str_v> ID STR REGEX
%token <int_v> INT
%token <bool_v> BOOL
%token <double_v> DOUBLE
%token AND OR LT GT EQ NE PF SF SS RE ANY
%token INTEGER_KW STRING_KW BOOLEAN_KW DOUBLE_KW
%token <nlin> IFCONFIG SELECT 
%token PRINTSTATS STATSONLY TIMERSTART TIMERSTOP 
%token OUTPUTON OUTPUTOFF OUTPUTTO OUTPUTCOUNT PREPROCESSLIMIT CLEAR CLEARRECYCLE CONSOLIDATE
/*
 * some semantic types for non-terminals
 */
%type <str_v>		Id
%type <value>		Value
%type <constraint>	Constraint
%type <constraint>	AnyConstraint
%type <message>		Message
%type <predicate>	Predicate
%type <op>		Op

%start StmtList

%%
/*
 * Grammar rules
 */
StmtList: /* empty */ | StmtList Stmt | StmtList error ';'
;

Stmt: IfConfig | Select 
 | PrintStats
 | CLEAR	{ FT->clear(); consolidate_guard = false; }
 | CLEARRECYCLE { FT->clear_recycle(); consolidate_guard = false; }
 | CONSOLIDATE  { 
     FT->consolidate(); 
     consolidate_guard = true;
}
 | STATSONLY	{ stats_only = true; }
 | TIMERSTART	{ sff_parser_timer_start (); }
 | TIMERSTOP	{ sff_parser_timer_stop (); }
 | OUTPUTON	{ output_level = SFF_VERBOSE; }
 | OUTPUTOFF	{ output_level = SFF_SILENT; }
 | OUTPUTCOUNT	{ output_level = SFF_MATCH_COUNT; }
 | OUTPUTTO STR
   { 
       sff_parser_open_output($2->c_str()); 
       delete($2);
   }
 | OUTPUTTO	{ sff_parser_open_output(0); }
 | PREPROCESSLIMIT EQ INT { 
     siena::FwdTable * FwdT = dynamic_cast<siena::FwdTable *>(FT);
     if (FwdT) {
	 FwdT->set_preprocess_rounds($3); 
     } else {
	 yyerror("Not using the default forwarding table.");
     }
 }
;

PrintStats:  PRINTSTATS	{ if (!errors) sff_parser_print_statistics(StatsFormat); }
 | PRINTSTATS STR 
   { 
       if (!errors) sff_parser_print_statistics(($2->length() > 0) ? ($2->c_str()) : StatsFormat); 
       delete($2);
   }
;

IfConfig: IFCONFIG INT Predicate OPT_Semicolon
   {
       if (consolidate_guard) {
	   std::cerr << yysfffname << ':' << $1 // implicitly nlin
		     << ": forwarding table already consolidated" << std::endl;
	   YYABORT;
       } 

       if (ifconfig_ranges && !(*ifconfig_ranges)[$2])
	   break;
       if (! stats_only) {
	   try {
	       FT->ifconfig($2, *$3);
	   } catch (siena::bad_constraint & ex) {
	       std::cerr << yysfffname << ':' << $1 // implicitly nlin
			 << ": bad constraint in predicate: " 
			 << ex.what() << std::endl;
	       delete($3);
	       YYABORT;
	   }
	   Mem.recycle();
	   delete($3);
       }
       ++i_counter;
   }
;

Id: ID { $$ = $1; }
 | INTEGER_KW { $$ = new std::string("integer"); }
 | STRING_KW { $$ = new std::string("string"); }
 | BOOLEAN_KW { $$ = new std::string("boolean"); }
 | DOUBLE_KW { $$ = new std::string("double"); }
 | IFCONFIG { $$ = new std::string("ifconfig"); }
 | SELECT { $$ = new std::string("select"); }
 | BOOL { $$ = ($1) ? (new std::string("true")) : (new std::string("false")); }
 | STR { $$ = $1; }
;

Predicate: Id Constraint 
   {
       if (! stats_only) {
	   $$ = new simple_predicate();
	   $$->add(new simple_filter());
	   $$->back()->add(new_sx_string($1), $2);
       }
       ++f_counter;
       ++c_counter;
       delete($1);
   }
 | Predicate AND Id Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->back()->add(new_sx_string($3), $4);
       }
       ++c_counter;
       delete($3);
   }
 | Predicate OR Id Constraint
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->add(new simple_filter());
	   $$->back()->add(new_sx_string($3), $4);
       }
       ++f_counter;
       ++c_counter;
       delete($3);
   }
;

Constraint: AnyConstraint { if (!stats_only) { $$ = $1; } }
 | Op INT 
   { 
       if (!stats_only) { $$ = new simple_op_value($1, $2); } 
   }
 | Op ANY 
   { 
       if (!stats_only) { $$ = new simple_op_value($1); } 
   }
 | Op DOUBLE 
   { 
       if (!stats_only) { $$ = new simple_op_value($1, $2); } 
   }
 | Op BOOL
   { 
       if (!stats_only) { $$ = new simple_op_value($1, $2); } 
   }
 | Op STR
   { 
       if (!stats_only) { $$ = new simple_op_value($1, new_sx_string($2)); }
       delete($2); 
   }
 | RE REGEX
   {
       if (!stats_only) { $$ = new simple_op_value(siena::re_id, 
						   new_sx_string($2)); }
       delete($2); 
   }
;

AnyConstraint: ANY INTEGER_KW
   { 
       if (!stats_only) { 
	   $$ = new simple_op_value(siena::any_id, (siena::int_t)0); 
       } 
   }
 | ANY ANY 
   { 
       if (!stats_only) { $$ = new simple_op_value(siena::any_id); } 
   }
 | ANY DOUBLE_KW
   { 
       if (!stats_only) { 
	   $$ = new simple_op_value(siena::any_id, (siena::double_t).0);
       } 
   }
 | ANY BOOLEAN_KW
   { 
       if (!stats_only) { 
	   $$ = new simple_op_value(siena::any_id, false);
       } 
   }
 | ANY STRING_KW
   { 
       if (!stats_only) { 
	   $$ = new simple_op_value(siena::any_id, new_sx_string()); 
       } 
   }
;

Op: LT { $$ = siena::lt_id; }
 |  GT { $$ = siena::gt_id; }
 |  EQ { $$ = siena::eq_id; }
 |  NE { $$ = siena::ne_id; }
 |  PF { $$ = siena::pf_id; }
 |  SF { $$ = siena::sf_id; }
 |  SS { $$ = siena::ss_id; }
;

Value: INT { if (!stats_only) { $$ = new simple_value($1); } }
 |  DOUBLE { if (!stats_only) { $$ = new simple_value($1); } }
 |  BOOL   { if (!stats_only) { $$ = new simple_value($1); } }
 |  ANY    { if (!stats_only) { $$ = new simple_value(); } }
 |  STR    { if (!stats_only) { $$ = new simple_value(new_sx_string($1)); }
             delete($1); } 
;

Select: SELECT Message OPT_Semicolon
   {
       if (!consolidate_guard) {
	   FT->consolidate();
	   consolidate_guard = true;
       } 
       if (! stats_only) {
	   switch (output_level) {
	   case SFF_MATCH_COUNT:
	   case SFF_SILENT: {
	       ICounter cc;
#ifndef WITH_STATISTICS
	       siena::sff_timer_stack.push(match_timer);
#endif
	       FT->match(*$2, cc);
#ifndef WITH_STATISTICS
	       siena::sff_timer_stack.pop();
#endif
	       if (cc.get_count() > 0) ++x_counter;
	       xx_counter += cc.get_count();
	       if (output_level == SFF_MATCH_COUNT) {
		   (*sff_output) << cc.get_count() << std::endl;
	       }
	   }
	   break;
	   case SFF_VERBOSE: {    
	       IPrinter ccc;
#ifndef WITH_STATISTICS
	       siena::sff_timer_stack.push(match_timer);
#endif
	       FT->match(*$2, ccc);
#ifndef WITH_STATISTICS
	       siena::sff_timer_stack.pop();
#endif
	       if (ccc.get_count() > 0) ++x_counter;
	       xx_counter += ccc.get_count();
	       ccc.flush(*sff_output);
	   }
	   }
	   Mem.recycle();
	   ++m_counter;
	   delete($2);
       }
   }
;

Message: Id EQ Value
   { 
       if (! stats_only) {
	   $$ = new simple_message();
	   $$->add(new_sx_string($1), $3);
       }
       ++a_counter;
       delete($1); 
   }
  | Message Id EQ Value
   {
       if (! stats_only) {
	   $$ = $1;
	   $$->add(new_sx_string($2), $4);
       }
       ++a_counter;
       delete($2); 
   }
;

OPT_Semicolon: ';' | /* empty */
;

%%

int yyerror(const char *s) 
{
    std::cerr << yysfffname << ':' << yylineno << ": " << s << std::endl;
    errors = true;
    stats_only = true;
    FT->clear();    
    return -1;
}

int sff_parser_run() throw()
{
    bool sync_state = ios_base::sync_with_stdio(false);
    siena::Timer::enable();
    sff_timer.start();
    if (!FT) FT = new siena::FwdTable();
    sff_parser_clear_statistics();
    siena::sff_timer_stack.push(parser_timer);
    int res = yyparse();
    siena::sff_timer_stack.pop();
    sff_timer.stop();
    siena::Timer::disable();
    sff_output->flush();
    ios_base::sync_with_stdio(sync_state);
    return res;
}

int sff_parser_stats_only() throw()
{
    stats_only = true;
    return sff_parser_run();
}

void sff_parser_output_off() throw()
{
    output_level = SFF_SILENT; 
}

void sff_parser_output_on() throw()
{
    output_level = SFF_VERBOSE; 
}

void sff_parser_output_level(sff_output_level_t l) throw()
{
    output_level = l;
}

void sff_parser_timer_start() throw()
{
    siena::Timer::enable();
    sff_timer.start();
}

void sff_parser_timer_stop() throw()
{
    sff_timer.stop();
    siena::Timer::disable();
}

void sff_parser_shutdown() throw()
{
    sff_parser_close_output();
    if (FT) delete(FT);
    Mem.clear();
}

void sff_parser_use_btable() throw()
{
    sff_parser_shutdown();
    FT = new siena::BTable();
}

void sff_parser_use_sorted_btable() throw()
{
    sff_parser_shutdown();
    FT = new siena::SortedBTable();
}

void sff_parser_use_bxtable() throw()
{
    sff_parser_shutdown();
    FT = new siena::BXTable();
}

void sff_parser_use_bctable() throw()
{
    sff_parser_shutdown();
    FT = new siena::BCTable();
}

void sff_parser_use_bvtable() throw()
{
    sff_parser_shutdown();
    FT = new siena::BVTable();
}

#ifdef HAVE_CUDD
void sff_parser_use_bddbtable() throw()
{
    sff_parser_shutdown();
    FT = new siena::BDDBTable();
}

void sff_parser_use_zddbtable() throw()
{
    sff_parser_shutdown();
    FT = new siena::ZDDBTable();
}
#endif

void sff_parser_use_fwdtable() throw()
{
    sff_parser_shutdown();
    FT = new siena::FwdTable();
}

void sff_parser_set_ifconfig_ranges(const char * ranges) throw()
{
    ifconfig_ranges = new siena::range_set<int> (ranges);
}

void sff_parser_set_statistics_format(const char * frmt) throw()
{
    StatsFormat = frmt;
}

void sff_parser_close_output() throw()
{
    if (sff_foutput) {
	sff_foutput->close();
	delete (sff_foutput);
	sff_foutput = 0;
	sff_output = & std::cout;
    }
}

int sff_parser_open_output(const char * filename) throw()
{
    sff_parser_close_output();
    if (filename) {
	sff_foutput = new std::ofstream(filename);
	if (sff_foutput->fail()) {
	    delete (sff_foutput);
	    sff_foutput = 0;
	    return -1;
	}
	sff_output = sff_foutput;
    } else {
	sff_output = & std::cout;
    }
    return 0;
}

void sff_parser_print_statistics(const char * format) throw()
{
    if (!format) 
	format = "i=%i f=%f c=%c n=%n a=%a m=%m M=%M s=%s\n";

    for(const char * f = format; *f != '\0'; ++f) {
	switch(*f) {
	case '%':
	    ++f;
	    switch(*f) {
	    case 'i': (*sff_output) << i_counter; break;
	    case 'f': (*sff_output) << f_counter; break;
	    case 'c': (*sff_output) << c_counter; break;
	    case 'n': (*sff_output) << m_counter; break;
	    case 'a': (*sff_output) << a_counter; break;
	    case 'm': (*sff_output) << x_counter; break;
	    case 'M': (*sff_output) << xx_counter; break;
	    case 's': (*sff_output) << FT->bytesize(); break;
	    case 'S': (*sff_output) << FT->allocated_bytesize(); break;
#ifdef WITH_STATISTICS
	    case 'b': (*sff_output) << siena::string_matcher_bytes; break;
#endif
	    case '%': (*sff_output) << '%'; break;
	    case 'T': {
		++f;
		long long time = 0;
		switch (*f) {
		case 't': time = sff_timer.read(); break;
		case 'p': time = parser_timer.read(); break;
#ifdef WITH_STATISTICS
		case 'I': time = siena::ifconfig_timer.read(); break;
		case 'C': time = siena::consolidate_timer.read(); break;
		case 'E': time = siena::bloom_encoding_timer.read(); break;
		case 'M': time = siena::match_timer.read(); break;
		case 'S': time = siena::string_match_timer.read(); break;
		case 'F': time = siena::forward_timer.read(); break;
#else
		case 'm': time = match_timer.read(); break;
#endif
		default:
		    std::cerr << "print_statistics: bad format character: '%T" 
			      << *f << "'" << std::endl;
		    if (*f == '\0') return;
		}
		(*sff_output) << time / 1000;
		break;
	    }
	    default:
		std::cerr << "print_statistics: bad format character: '%" 
			  << *f << "'" << std::endl;
		if (*f == '\0') return;
	    }
	    break;

	case '\\': 		
	    ++f;
	    switch(*f) {
	    case 'n': (*sff_output) << '\n'; break;
	    case 't': (*sff_output) << '\t'; break;
	    case 'a': (*sff_output) << '\a'; break;
	    case 'v': (*sff_output) << '\v'; break;
	    case '\\': (*sff_output) << '\\'; break;
	    default:
		std::cerr << "print_statistics: bad escape character: '" 
			  << *f << "'" << std::endl;
		if (*f == '\0') return;
	    }
	    break;
	default:
	    (*sff_output) << *f;
	}
    }
    sff_output->flush();
}

void sff_parser_clear_statistics() throw()
{
    i_counter = 0;
    f_counter = 0;
    c_counter = 0;
    m_counter = 0;
    a_counter = 0;
    x_counter = 0;
    xx_counter = 0;
    sff_timer.reset();
    parser_timer.reset();
#ifdef WITH_STATISTICS
    siena::string_matcher_bytes = 0;
    siena::ifconfig_timer.reset();
    siena::consolidate_timer.reset();
    siena::match_timer.reset();
    siena::bloom_encoding_timer.reset();
    siena::string_match_timer.reset();
#endif
}
