/*
 *  This file is part of Siena, a wide-area event notification system.
 *  See http://www.inf.usi.ch/carzaniga/siena/
 *  
 *  Author: Antonio Carzaniga
 *  See the file AUTHORS for full details. 
 *  
 *  Copyright (C) 1998-2003 University of Colorado
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA, or send email to one of the authors.
 *  
 *  
 *  $Id: sff_scanner.ll,v 1.21 2010-03-11 15:16:41 carzanig Exp $
 */
%option yylineno
%option interactive
%option always-interactive
%{
#include <string>
#include <cstdlib>

#include "yysfftypes.h"
/* "yysfftypes.h" must be #included before "sff_parser.h" */
#include "sff_parser.h"

#define YY_NO_UNPUT

std::string yysfffname = "{standard input}"; /* Current source file name */
extern int yyerror (const char *);

int sff_scanner_open(const char * fname) throw() {
    if (strcmp (fname, "-")) {
        yysfffname = fname;
        yyin = fopen(fname, "r");
    } else {
	yysfffname = "{standard input}";
        yyin = stdin;
    }
    yylineno = 1;
    return yyin == NULL;
}

int sff_scanner_close() throw() {
    if (yyin != NULL && yyin != stdin) {
	int res = fclose(yyin);
	yyin = NULL;
	return res;
    } else 
	return 0;
}

#define RET(x) {yylval->nlin = yylineno; yylval->fname = & yysfffname;};return(x)

%}

space			[ \b\t\v\f\r]+
alpha			[a-zA-Z_]
nonzero			[1-9]
digit			[0-9]
oct_digit		[0-7]
hex_digit		[0-9a-fA-F]
ident_extra		[/.\-]
ident			{alpha}({digit}|{alpha}|{ident_extra})*
sign_opt		[\-+]?
decimal			({sign_opt}{nonzero}{digit}*)|("0")
octal			{oct_digit}{1,3}
strhex			"0"?[Xx]{hex_digit}+
hex			"0"[Xx]{hex_digit}+
exponent		[eE]{sign_opt}{digit}+
fractional		({digit}+".")|({digit}*"."{digit}+)
fpoint			{sign_opt}{fractional}{exponent}?
u			[Uu]
l			[Ll]
f			[Ff]

/* Parts of a regular expressions. not_bracket is simply "everything but
   a bracket", but flex versions > 2.5.4 do not like two consecutive
   brackets in the input file!

   When parsing a regex, we don't want the / to be a terminator if
   it is within a bracket, i.e. /[abc/def]/ must be a valid regex.
   This is like sed, but unlike perl.  For this reason, we define one
   start conditions for brackets, that eats a 'bracket_elem' at a time.
   Note that a bracket is started with "["\^?\]? in order to eat the
   initial caret and a literal closing bracket.

   Also note that we want to parse escapes differently in strings and
   regexes, which is why we need a different token:

   0) Of course the above handling of brackets does not belong in strings.
   1) \b (bell, ASCII 7) is not supported in regexes.
   2) Unrecognized escapes (e.g. \< or \b itself) must be passed down
      to the regex matcher, rather than stripped. 
   3) We give an error if a regex is multiline unless each line is
      \-terminated.  In the future this may be extended to strings.

   coll_class is something like Perl's \[: .*? :\] and it can be
   "dissected" this way:

	"[:"			begin the sequence
	(\]|			closing bracket not preceded by :
	\:{not_bracket}|	or : not followed by closing bracket
	[^\]:])*		or any other char -- ad libitum
	":"+\]			and the end of the sequence		*/

not_bracket		(a|[^\]a])
coll_class		"[:"(\]|\:{not_bracket}|[^\]:])*":"+\]
coll_elem		"[."(\]|\.{not_bracket}|[^\].])*"."+\]
coll_equiv		"[="(\]|\={not_bracket}|[^\]=])*"="+\]
bracket_elem		({not_bracket}|{coll_elem}|{coll_class}|{coll_equiv})

/* 
 *  mutually exclusive states 
 */
%x comment strlit regex bracket

%% 

^#.*\n			{ const char * nlin; const char * fname;
                          if ((nlin = strtok(yytext, "\n#\t")) != NULL &&
                              (fname = strtok(NULL, " \t\"")) != NULL) {
			      yylineno = strtol(nlin, NULL, 10);
			      yysfffname = fname;
                          }
                        }

"/*"			BEGIN(comment);
<comment>[^\*\n]*	;
<comment>\*[\*]*[^/\n]	;
<comment>\n		;
<comment>[\*]*"*/"	BEGIN(INITIAL);

"//"[^\n]*\n		;		

{space}			; 
\n			;

\"			{ yylval->str_v = new std::string(); BEGIN(strlit); }
"/"			{ yylval->str_v = new std::string(); BEGIN(regex); }

<strlit,regex,bracket>{
  /* More specific rules come first, so that they win over the generic "\\."
     rule below.  */
  \\n		*yylval->str_v += '\n';
  \\t		*yylval->str_v += '\t';
  \\r		*yylval->str_v += '\r';
  \\e		*yylval->str_v += '\e';
  \\f		*yylval->str_v += '\f';
  \\a		*yylval->str_v += '\a';
  \\v		*yylval->str_v += '\v';
  \\{octal}	{ char *c = yytext + 1; /* skip '\\' */
                  *yylval->str_v += (char)(strtoul(c, NULL, 8) % 256); }
  \\{strhex}	{ char *c = yytext + (yytext[1] == '0' ? 3 : 2); /* skip "\\0x" */
                  *yylval->str_v += (char)(strtoul(c, NULL, 16) % 256); }
}

<strlit>{
  \\b		*yylval->str_v += '\b';
  \n		{ *yylval->str_v += '\n'; }
  \"		{ BEGIN(INITIAL); RET(STR); }
  [^\"\\\n]+	{ yylval->str_v->append(yytext); }
  \\\?		*yylval->str_v += '\?';
  \\.		*yylval->str_v += yytext[1];
}

<regex,bracket>{
  /* A regex is very similar to a string literal, but unrecognized escapes
     are passed to the regex parser for further interpretation.  In addition,
     \cX sequences are recognized to mean Ctrl-X.  */
  \\c[a-z]		    *yylval->str_v += (char) (yytext[2] ^ 96);
  \\c[?-_]		    *yylval->str_v += (char) (yytext[2] ^ 64);
  \\.			    { yylval->str_v->append(yytext); }
  \n			    { yyerror ("new-line in regex"); BEGIN(INITIAL);
			      RET (REGEX); }
}

<regex>{
  "/"			    { BEGIN(INITIAL); RET(REGEX); }
  [^[/\\\n]+		    { yylval->str_v->append(yytext); }
  \[\^?\]?		    { yylval->str_v->append(yytext); BEGIN (bracket); }
}

<bracket>{
  \]			    { yylval->str_v->append(yytext); BEGIN(regex); }
  {bracket_elem}	    { yylval->str_v->append(yytext); }
}

","			RET(AND);
"|"			RET(OR);
"&&"			RET(AND);
"||"			RET(OR);
"="			RET(EQ);
">"			RET(GT);
"<"			RET(LT);
"!="			RET(NE);
"=*"			RET(PF);
"*="			RET(SF);
"**"			RET(SS);
"~="			RET(RE);

"ifconfig"		RET(IFCONFIG);
"select"		RET(SELECT);

"%print_statistics"	RET(PRINTSTATS);
"%statistics_only"	RET(STATSONLY);
"%timer_start"		RET(TIMERSTART);
"%timer_stop"		RET(TIMERSTOP);
"%output_on"		RET(OUTPUTON);
"%output_off"		RET(OUTPUTOFF);
"%output_counter"	RET(OUTPUTCOUNT);
"%output_to"		RET(OUTPUTTO);
"%pre_process_limit"	RET(PREPROCESSLIMIT);
"%clear"		RET(CLEAR);
"%clear_recycle"	RET(CLEARRECYCLE);
"%consolidate"		RET(CONSOLIDATE);

"true"			{ yylval->bool_v = true; RET(BOOL);}
"false"			{ yylval->bool_v = false; RET(BOOL);}

"any"			RET(ANY);
"integer"		RET(INTEGER_KW);
"string"		RET(STRING_KW);
"boolean"		RET(BOOLEAN_KW);
"double"		RET(DOUBLE_KW);

{ident}			{ yylval->str_v = new std::string(yytext); RET(ID); }

{decimal}		{ yylval->int_v = strtoll(yytext, NULL, 10); RET(INT); }

{hex}			{ yylval->int_v = strtoul(yytext, NULL, 16); RET(INT); }

{fpoint}		{ yylval->double_v = strtod(yytext, NULL); RET(DOUBLE); }

.			RET(*yytext);

%%

int yywrap() { return 1; }
